/*-
 * Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: dimentia.c,v 1.23 2024/06/15 16:22:29 babolo Exp $"

#define BLIN_COMPAT      4
#define DIMENT_INTERNAL

#include <sys/types.h>
#include <sysexits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <malloc_np.h>
#include <babolo/BLINflag.h>
#include "dimentia.h"

diment_desc *
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_init(BLIN_flag flags, uint unit, int alnmem, uint lemax, const int *size) {                 /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (flags & BLIN_MASK)
    diment_desc *diment = NULL;
    size_t       mds;
    int32_t      szdown = 0;
    int32_t      szup = 1;

    ifBLIN_QX2("+ flags=%08X level=%d unit=%d", flags, lemax, unit);
    
    if  (!!(flags & DIMENT_BYTES)) {
        unit *= 8;
        alnmem += 3;
    }
    if  (lemax < 1 || lemax > 30) {
        errno = EINVAL;
        ifBLIN_QW0("level: %d", lemax);
        goto errout;
    }
    if  (unit == 0) {
        errno = EINVAL;
        ifBLIN_QW0("unit: %d", unit);
        goto errout;
    }
    if  (alnmem > 43) {
        errno = EINVAL;
        ifBLIN_QW0("alnmem > 43");
        goto errout;
    }
    mds = lemax * sizeof(diment->size[0]) + sizeof(diment_desc);
    if  (!(diment = malloc(mds))) {
        ifBLIN_QW0("No mem");
        goto errout;
    }
    bzero(diment, mds);
    diment->flags  = flags;
    diment->level  = lemax;
    diment->unit   = unit;
    diment->alnmem = alnmem;
    diment->fill.xdown = POW2(8 * sizeof(diment->fill.xdown) - 1) - 1;
    diment->fill.xup   = POW2(8 * sizeof(diment->fill.xup  ) - 1)    ;
    diment->fill.ydown = POW2(8 * sizeof(diment->fill.ydown) - 1) - 1;
    diment->fill.yup   = POW2(8 * sizeof(diment->fill.yup  ) - 1)    ;
    diment->fl = calloc(1, DIMENT_b2B(diment->unit));
    if  (!diment->fl) {
        ifBLIN_QW0("No mem");
        goto errout;
    }
    diment->alnkoef = DIMENT_BITBIGCEIL(unit, alnmem);
    if  (diment->alnkoef < unit) {
        errno = EDOOFUS;
        ifBLIN_QX0("WRONG MATH");
        goto errout;
    }
    if  (diment->alnkoef % 8 != 0) { /* XXXX */
        errno = ENOEXEC;
        ifBLIN_QX0("bit alignment is not awailable");
        goto errout;
    }
    if  (!size) {
        errno = EINVAL;
        ifBLIN_QW0("size==NULL");
        goto errout;
    } else {
        // size[0] == 0 is valid (tile of size 1)
        mds = size[0];
        for (int i = 1; i < diment->level; i++) {
            if  (size[i] < 1) {
                errno = EINVAL;
                ifBLIN_QW0("size[%d] < 1", i);
                goto errout;
            }
            mds += size[i];
        }
        if  (mds > 30) {/*       */
            errno = EINVAL;
            ifBLIN_QW0("size > 30");
            goto errout;
        }
        for (uint i = 0; i < lemax; i++) {
            diment->size[i] = size[i];
        }
    }
    for (int i = 0; i < diment->level; i++) {
        szup <<= diment->size[i];
    }
    if  (!(diment->flags & DIMENT_NONEG)) {
        szup >>= 1;
        szdown = -szup;
    }
    szup -= 1;
    diment->margin.xdown = szdown;
    diment->margin.xup   = szup  ;
    diment->margin.ydown = szdown;
    diment->margin.yup   = szup  ;
    ifBLIN_QX2( "- flags=%08X level=%u unit=%u size=%"BLIN_U
              , diment->flags, diment->level, diment->unit, mds
              );
    return(diment);
errout:
    free(diment);
    return(NULL);
#   undef blin_internal_flags
}

void
/*****************************************************************************************************
 **      level. 0 -                                        **/
indir_fini(diment_desc *diment, diment_inlar* layer, uint level) {                                 /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    ifBLIN_QX4("+%u %u %"BLIN_O"u", level, diment->size[level], layer->index);
    if  (!diment) goto out;
    if  (!layer) goto out;
    if  (!layer->term) goto out;
    if  (level > 0) {
        for (u_int64_t i = 0; i < POW2(diment->size[level] * 2); i++) {
            ifBLIN_QX4("i%u %"BLIN_O"d", level, i);
            indir_fini(diment, &(layer->next[i]), level - 1);
        }
    }
    if  (!!diment->tofree) {
        for (u_int64_t i = 0; i < POW2(diment->size[0] * 2); i++) {
            ifBLIN_QX4("%"BLIN_O"d", i);
            diment->tofree(&layer->term[i]);
    }   }
    free(layer->term);
    ifBLIN_QX4("-%u done", level);
    return;
out:
    ifBLIN_QX4("-%u", level);
    return;
#   undef blin_internal_flags
}

void
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_fini(diment_desc *diment) {                                                                 /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (!diment) return;
    indir_fini(diment, &diment->top, diment->level - 1);
    free(diment->fl);
    free(diment);
}

int
/*****************************************************************************************************
 **       level. 0 -                                                    **/
indir_init(diment_desc *diment, diment_inlar *layer, uint level) {                                 /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int       ex = 0;
    
    ifBLIN_QX4("+ %u %"BLIN_X" %"BLIN_X, level, BLIN_I(layer), BLIN_I(diment->fl));
    if  (!diment) {
        ifBLIN_QX0("diment==NULL");
        ex = EINVAL;
        goto out;
    }
    if  (!layer) {
        ifBLIN_QX0("layer==NULL");
        ex = EINVAL;
        goto out;
    }
    if  (level > 0) {
        layer->index = POW2(diment->size[level] * 2);
        layer->next = calloc(POW2(diment->size[level] * 2), sizeof(diment_inlar));
        ifBLIN_QX6("usable 0x%zX", malloc_usable_size(layer->next));
        if  (!layer->next) {
            ifBLIN_QW0("no mem");
            ex = errno;
            goto out;
        }
    } else {
        if  (diment->alnkoef % 8 == 0) {
            layer->index = POW2(diment->size[0] * 2) * (diment->alnkoef / 8);
            layer->term = malloc(POW2(diment->size[0] * 2) * (diment->alnkoef / 8));
            ifBLIN_QX6("usable 0x%zX", malloc_usable_size(layer->term));
            if  (!layer->term) {
                ifBLIN_QW0("no mem");
                ex = errno;
                goto out;
            }
            for (u_int64_t i = 0; i < POW2(diment->size[0] * 2); i++) {
                memcpy(layer->term + i * diment->alnkoef / 8, diment->fl, DIMENT_b2B(diment->unit));
            }
        } else {
            /* XXXX */
            ifBLIN_QX0("UNAVAILABLE");
            ex = ENOEXEC;
            goto out;
        }
    }
out:
    ifBLIN_QX4("-");
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_setfl(diment_desc *diment, void *fl) {                                                      /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    
    if  (!diment || !fl) return(ENOENT);
    memcpy(diment->fl, fl, DIMENT_b2B(diment->unit));
    return(0);
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_getfl(diment_desc *diment, void *fl) {                                                      /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    
    if  (!diment || !fl) return(ENOENT);
    memcpy(fl, diment->fl, DIMENT_b2B(diment->unit));
    return(0);
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                        **/
diment_setmargin(diment_desc *diment, diment_bords b) {                                            /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    int       ex = 0;
    int32_t   szdown = 0;
    int32_t   szup = 1;
    
    if  (!diment) return(ENOENT);
    if  (b.xdown > b.xup) {
        b.xdown ^= b.xup  ;
        b.xup   ^= b.xdown;
        b.xdown ^= b.xup  ;
        ex = EINVAL;
    }
    if  (b.ydown > b.yup) {
        b.ydown ^= b.yup  ;
        b.yup   ^= b.ydown;
        b.ydown ^= b.yup  ;
        ex = EINVAL;
    }
    if  (!!(diment->flags & DIMENT_BEYOND)) {
        diment->margin = b;
        return(ex);
    }
    for (int i = 0; i < diment->level; i++) {
        szup <<= diment->size[i];
    }
    if  (!(diment->flags & DIMENT_NONEG)) {
        szup >>= 1;
        szdown = -szup;
    }
    szup -= 1;
    diment->margin.xdown = MAX(b.xdown, szdown);
    diment->margin.xup   = MIN(b.xup  , szup  );
    diment->margin.ydown = MAX(b.ydown, szdown);
    diment->margin.yup   = MIN(b.yup  , szup  );
    if  (b.xdown < szdown) ex = EDOM;
    if  (b.xup   > szup  ) ex = EDOM;
    if  (b.ydown < szdown) ex = EDOM;
    if  (b.yup   > szup  ) ex = EDOM;
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_getmargin(diment_desc *diment, diment_bords *b) {                                           /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (!diment || !b) return(ENOENT);
    *b = diment->margin;
    return(0);
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_getfill(diment_desc *diment, diment_bords *b) {                                             /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
    if  (!diment) return(ENOENT);
    if  (!b) return(ENOENT);
    *b = diment->fill;
    return(0);
}

void
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                **/
diment_setconv( diment_desc *diment                                                                /**
 **/          , void *mem                                                                          /**
 **/          , int (*conv) (void*, int32_t*, int32_t*)                                            /**
 **/          , int (*fconv)(void*, int32_t*, int32_t*, double, double)                            /**
 **/          ) {                                                                                  /**
 *****************************************************************************************************
 *****************************************************************************************************/
    diment->convmem = mem;
    diment->conv    = conv;
    diment->fconv   = fconv;
}
