/*-
 * Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: dimentia.h,v 1.24 2024/09/24 11:53:50 cuhegh Exp $
 */

#ifndef _DIMENTIA_H_
# define _DIMENTIA_H_

# ifndef _SYS_TYPES_H_
#  warning this file needs sys/types.h as a prerequisite
#  include <sys/types.h>
# endif

# ifndef _BABOLO_BLINflag_H_
#  warning this file needs babolo/BLINflag.h as a prerequisite
#  include <babolo/BLINflag.h>
# endif

__BEGIN_DECLS

typedef struct diment_inlar diment_inlar;

struct diment_inlar {
    union {
        diment_inlar *next;
        void         *term;
    };               /*                 */
    u_int64_t  index;    /*   + 1      */
};

typedef struct diment_bords {
    int32_t xdown;
    int32_t xup  ;
    int32_t ydown;
    int32_t yup  ;
} diment_bords;

typedef struct {
    BLIN_flag   flags;               /*    */
/*         BLIN_GEN7     0xFF000000       */
#   define DIMENT_NONEG    0x800000  /*   (0 0)        */
                                     /*   DIMENT_BEYOND      */
#   define DIMENT_BEYOND   0x400000  /*   margin    */
#   define DIMENT_TKNENT   0x200000  /*                    */
#   define DIMENT_CREOUT   0x100000  /*        */
#   define DIMENT_BYTES    0x080000  /*   .     */
#   define DIMENT_NOBG     0x000000  /* XXXX     bg. */
#   define DIMENT_FLAGMASK 0xFF0000  /*   */
#   define DIMENT_VAR_MASK ( BLIN_GEN7    | DIMENT_NONEG  | DIMENT_BEYOND  \
                           | DIMENT_TKNENT | DIMENT_CREOUT | DIMENT_BYTES  \
                           )         /*       */
    u_int16_t    unit;            /*                                           */
    int8_t       alnmem;          /*   .   2 ** n,  */
    u_int8_t     level;           /*                              */
    diment_inlar top;
    void        *fl;              /*     */
    void       (*tofree)(void*);  /* XXXX */
    void        *dummy;
    void        *convmem;
    int        (*conv) (void*, int32_t*, int32_t*);
    int        (*fconv)(void*, int32_t*, int32_t*, double, double);
    diment_bords margin;          /*   */
    diment_bords fill;            /*   */
    u_int64_t    alnkoef;         /* Σ ,  */
    int8_t       size[0];         /*          0-  */
} diment_desc;

# define DIMENT_b2B(A) (((A) + 7) / 8)
# define DIMENT_FLAGS(DIMENT)         ((DIMENT)->flags | 0) // to not allow assining
# define diment_tl_cr(DIMENT, X, Y)   diment_tile(DIMENT, X, Y, 1)
# define diment_tl_tk(DIMENT, X, Y)   diment_tile(DIMENT, X, Y, 0)
# define diment_doflag(DIMENT, W, MD) blin_doflag(MD, W, DIMENT_VAR_MASK, &(DIMENT)->flags)
# define DIMENT_TILE_INDEX(DIMENT, X, Y) ({ \
    u_int64_t size = (DIMENT)->size[0]; \
    (((Y) & MASK(size)) << size) | ((X) & MASK(size)); \
}) /*        (   ) */

diment_desc *diment_init      __P(( BLIN_flag flags
                                  , uint unit
                                  , int alnmem
                                  , uint lemax
                                  , const int *size
                                 ));
void         diment_fini      __P((diment_desc *));
int          diment_setfl     __P((diment_desc *, void *fl));
int          diment_getfl     __P((diment_desc *, void *fl));
int          diment_setmargin __P((diment_desc *, diment_bords b));
int          diment_getmargin __P((diment_desc *, diment_bords *b));
int          diment_getfill   __P((diment_desc *, diment_bords *b));
void         diment_setconv   __P(( diment_desc *
                                  , void *mem
                                  , int (*conv) (void*, int32_t*, int32_t*)
                                  , int (*fconv)(void*, int32_t*, int32_t*, double, double)
                                 ));
int          diment_donut     __P((void *diment, int32_t *x, int32_t *y));
int          diment_unit      __P(( diment_desc *
                                  , diment_inlar *
                                  , int32_t x
                                  , int32_t y
                                  , void *uout
                                  , void *uin
                                 ));
diment_inlar *diment_tile     __P((diment_desc *, int32_t x, int32_t y, int do_init));
int           diment_un_cr    __P((diment_desc *, int32_t x, int32_t y, void *unit));
int           diment_un_tk    __P((diment_desc *, int32_t x, int32_t y, void *unit));
int           diment_iun_cr   __P((diment_desc *, int32_t x, int32_t y, void *unit));
int           diment_iun_tk   __P((diment_desc *, int32_t x, int32_t y, void *unit));
int           diment_fun_cr   __P((diment_desc *, double  x, double  y, void *unit));
int           diment_fun_tk   __P((diment_desc *, double  x, double  y, void *unit));
int           diment_flush    __P((diment_desc *diment, diment_bords br, int fd, int alnout));

#ifdef DIMENT_INTERNAL

# define MAX(A, B) ((A) > (B) ? (A) : (B))
# define MIN(A, B) ((A) < (B) ? (A) : (B))
# define ABS(A)    ((A) >= 0  ? (A) : -(A))
# define POW2(A)   ((u_int64_t)1 << (u_int64_t)(A))
# define MASK(B)   (POW2(B) - 1)
# define DIMENT_BITBIGCEIL(A, BIT) (((A) + MASK(BIT)) & ~MASK(BIT))
# define DIMENT_ZEROS256 ("\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                          "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0" \
                         ) /* 256 of '\0' */

int           indir_init    __P((diment_desc *, diment_inlar *, uint level));
void          indir_fini    __P((diment_desc *, diment_inlar *, uint level));
int           deco          __P((diment_desc *, int32_t idx, u_int64_t *ptx));

#endif

__END_DECLS

#endif /* !_DIMENTIA_H_ */
