/*-
 * Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: pixel.c,v 1.11 2024/09/24 11:53:50 cuhegh Exp $"

#define BLIN_COMPAT      4
#define DIMENT_INTERNAL

#include <sys/types.h>
#include <sysexits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <babolo/BLINflag.h>
#include "dimentia.h"

int
/*****************************************************************************************************
 **                                                  **/
deco(diment_desc *diment, int32_t cord, u_int64_t *ptx) {                                          /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int ex = 0;
    
    ifBLIN_QX4("+ %d", cord);
    if  (!diment || !ptx) {
        ifBLIN_QX0("<pointer>==NULL");
        ex = EINVAL;
        goto out;
    }
    ifBLIN_QX6("checked");
    for (int i = 0; i < diment->level; i++) {
        ptx[i] = cord & MASK(diment->size[i]);
        ifBLIN_QX6("ptx[%d]==%"BLIN_O"d", i, ptx[i]);
        cord >>= diment->size[i];
    }
    if  (!(diment->flags & DIMENT_NONEG)) {
        ptx[diment->level - 1] += POW2(diment->size[diment->level - 1]  - 1);
        ptx[diment->level - 1] &= POW2(diment->size[diment->level - 1]) - 1 ;
    }
out:
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

diment_inlar *
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                      **/
diment_tile(diment_desc *diment, int32_t x, int32_t y, int do_init) {                              /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    diment_inlar *layer = NULL;
    u_int64_t    xp[30];
    u_int64_t    yp[30];

    ifBLIN_QX3("+");
    if  (!diment) {
        errno = EINVAL;
        ifBLIN_QW0("diment==NULL");
        goto out;
    }
    ifBLIN_QX5("checked");
    deco(diment, x, xp);
    deco(diment, y, yp);
    ifBLIN_QX5("decoded");
    layer = &diment->top;
    for (int lev = diment->level - 1; lev > 0; lev--) {
        if  (!layer->next) {
            if  (!do_init) {
                layer = NULL;
                errno = ENOENT;
                goto out;
            }
            if  (!!(errno = indir_init(diment, layer, lev))) {
                layer = NULL;
                goto out;
        }   }
        yp[lev] <<= diment->size[lev];
        ifBLIN_QX5("layer index %"BLIN_O"d] 0x%"BLIN_O"d", xp[lev] + yp[lev], layer->index);
        layer = &layer->next[xp[lev] + yp[lev]];
    }
    if  (!layer->term) {
        if  (!do_init) {
            layer = NULL;
            errno = ENOENT;
            goto out;
        }
        if  (!!(errno = indir_init(diment, layer, 0))) {
            layer = NULL;
            goto out;
    }   }
out:
    ifBLIN_QX3("- %"BLIN_X" %d", BLIN_I(layer), errno);
    return(layer);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **  /                                                                         **/
diment_unit(diment_desc *diment, diment_inlar *lar, int32_t x, int32_t y, void *uout, void *uin) { /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int           ex = 0;
    void         *px = NULL;   /*  */
    u_int64_t     indx;

    ifBLIN_QX3("+ %d %d", x, y);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (  (x < diment->margin.xdown)
        || (x > diment->margin.xup  )
        || (y < diment->margin.ydown)
        || (y > diment->margin.yup  )
        ) {
        ifBLIN_QX3("out of margin");
        ex = EDOM;
        goto out;
    }
    if  (!uin && !uout) goto out;
    if  (!lar) {
        ex = EINVAL;
        ifBLIN_QW0("layer == NULL");
        goto out;
    }
    ifBLIN_QX5("checked");
    indx = DIMENT_TILE_INDEX(diment, x, y);
    ifBLIN_QX5("layer index %"BLIN_O"u 0x%"BLIN_O"X", indx, lar->index);
    
    px = lar->term;
    if  (diment->alnkoef % 8 == 0) {
        px += indx * diment->alnkoef / 8;
        if  (!!uout) {
            memcpy(uout, px, DIMENT_b2B(diment->unit));
        }
        if  (!!uin) {
            memcpy(px, uin, DIMENT_b2B(diment->unit));
        }
    } else {
        ex = ENOEXEC; /* XXXX */
        ifBLIN_QW0("UNAVAILABLE");
        goto out;
    }
    if  (!!uin) {
        if  (x < diment->fill.xdown) diment->fill.xdown = x;
        if  (x > diment->fill.xup  ) diment->fill.xup   = x;
        if  (y < diment->fill.ydown) diment->fill.ydown = y;
        if  (y > diment->fill.yup  ) diment->fill.yup   = y;
    }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_un_cr(diment_desc *diment, int32_t x, int32_t y, void *unit) {                              /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int           ex = 0;
    diment_inlar *layer = NULL;
    
    ifBLIN_QX3("+ %d %d", x, y);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (!unit) {
        ex = EINVAL;
        ifBLIN_QX0("unit == NULL");
        goto out;
    }
    ex = diment_unit(diment, NULL, x, y, NULL, NULL);
    if  (!!ex) {
        if  (!(diment->flags & DIMENT_CREOUT)) {
            ex = 0;
        }
        goto out;
    }
    ifBLIN_QX5("checked");
    layer = diment_tile(diment, x, y, 1);
    if  (!layer) {
        ifBLIN_QW0("diment_tile");
        ex = errno;
        goto out;
    }
    ex = diment_unit(diment, layer, x, y, NULL, unit);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_un_tk(diment_desc *diment, int32_t x, int32_t y, void *unit) {                              /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int           ex = 0;
    diment_inlar *layer = NULL;
    
    ifBLIN_QX3("+ %08x %d %d", diment->flags, x, y);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (!unit) {
        ex = EINVAL;
        ifBLIN_QX0("unit == NULL");
        goto out;
    }
    ex = diment_unit(diment, NULL, x, y, NULL, NULL);
    if  (!!ex) {
        ifBLIN_QX3("out of margin");
        goto out;
    }
    ifBLIN_QX5("checked");
    layer = diment_tile(diment, x, y, 0);
    if  (!layer) {
        if  (errno == ENOENT) {
            if  (!(diment->flags & DIMENT_TKNENT)) {
                errno = 0;
                memcpy(unit, diment->fl, DIMENT_b2B(diment->unit));
            } else {
                ex = ENOENT;
            }
            goto out;
        }
        ifBLIN_QW0("diment_tile");
        ex = errno;
        goto out;
    }
    ex = diment_unit(diment, layer, x, y, unit, NULL);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_iun_cr(diment_desc *diment, int32_t x, int32_t y, void *unit) {                             /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int   ex = 0;
    
    ifBLIN_QX3("+ %08x %d %d", diment->flags, x, y);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (!unit) {
        ex = EINVAL;
        ifBLIN_QX0("unit == NULL");
        goto out;
    }
    if  (!!diment->conv) {
        ex = diment->conv(diment->convmem, &x, &y);
        if  (!!ex) {
            ifBLIN_QX0("convercion failed %d", ex);
            goto out;
    }   }
    ex = diment_un_cr(diment, x, y, unit);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_fun_cr(diment_desc *diment, double fx, double fy, void *unit) {                             /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int   ex = 0;
    int32_t x, y;
    
    ifBLIN_QX3("+ %08x %f %f", diment->flags, fx, fy);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (!unit) {
        ex = EINVAL;
        ifBLIN_QX0("unit == NULL");
        goto out;
    }
    if  (!!diment->fconv) {
        ex = diment->fconv(diment->convmem, &x, &y, fx, fy);
        if  (!!ex) {
            ifBLIN_QX0("convercion failed %d", ex);
            goto out;
        }
    } else {
        x = lroundf(fx);
        y = lroundf(fy);
    }
    ex = diment_un_cr(diment, x, y, unit);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_iun_tk(diment_desc *diment, int32_t x, int32_t y, void *unit) {                             /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int       ex = 0;
    
    ifBLIN_QX3("+ %08x %d %d", diment->flags, x, y);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (!unit) {
        ex = EINVAL;
        ifBLIN_QX0("unit == NULL");
        goto out;
    }
    if  (!!diment->conv) {
        ex = diment->conv(diment->convmem, &x, &y);
        if  (!!ex) {
            ifBLIN_QX0("convercion failed %d", ex);
            goto out;
    }   }
    ex = diment_un_tk(diment, x, y, unit);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 **                                                                                                 **/
diment_fun_tk(diment_desc *diment, double fx, double fy, void *unit) {                             /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    int   ex = 0;
    int32_t x, y;
    
    ifBLIN_QX3("+ %08x %f %f", diment->flags, fx, fy);
    if  (!diment) {
        ex = EINVAL;
        ifBLIN_QX0("diment == NULL");
        goto out;
    }
    if  (!unit) {
        ex = EINVAL;
        ifBLIN_QX0("unit == NULL");
        goto out;
    }
    if  (!!diment->fconv) {
        ex = diment->fconv(diment->convmem, &x, &y, fx, fy);
        if  (!!ex) {
            ifBLIN_QX0("convercion failed %d", ex);
            goto out;
        }
    } else {
        x = lroundf(fx);
        y = lroundf(fy);
    }
    ex = diment_un_tk(diment, x, y, unit);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}















/*  diment_setconv(diment, diment, diment_donut, NULL) */
int
/*****************************************************************************************************
 *****************************************************************************************************
 **    .    .                              **/
diment_donut(void *mem, int32_t *x, int32_t *y) {                                                  /**
 **                                                                                                 **
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    diment_desc *diment = mem;
    int      ex = EX_OK;
    int32_t  marginsize;
    
    ifBLIN_QX4("+");
    if  (!diment) {
        ifBLIN_QX0("diment==NULL");
        ex = EX_USAGE;
        goto out;
    }
    if  (!!x) {
        marginsize = diment->margin.xup - diment->margin.xdown + 1;
        *x -= diment->margin.xdown;
        *x = (*x % marginsize + marginsize) % marginsize;
        *x += diment->margin.xdown;
    }
    if  (!!y) {
        marginsize = diment->margin.yup - diment->margin.ydown + 1;
        *y -= diment->margin.ydown;
        *y = (*y % marginsize + marginsize) % marginsize;
        *y += diment->margin.ydown;
    }
out:
    ifBLIN_QX4("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}
