/*-
 * Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: psout.c,v 1.11 2024/02/10 04:02:28 cuhegh Exp $"

#define BLIN_COMPAT      4
#define DIMENT_INTERNAL

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include "dimentia.h"

int
/*****************************************************************************************************
 **                                                                          **/
diment_flush(diment_desc *diment, diment_bords br, int fd, int alnout) {                           /**
 **                                                                                                 **
 *****************************************************************************************************/
#   define isJX(EXCODE, FORMAT, ...) \
        do { ifBLIN_QX0(FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define isJW(EXCODE, FORMAT, ...) \
        do { ifBLIN_QW0(FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    u_char    buf[256];
    uint      bufsz = 0;      /* bits                     */
    int       printed = 0;    /*              */
    int       ex = 0;         /*               */
    u_int64_t psize;          /* bytes                    */
    u_int64_t wsize;          /*   bits */
    u_char   *px = calloc(1, DIMENT_b2B(diment->unit) + 1);
    BLIN_flag tmpfl;
    
    ifBLIN_QX3("+ %d %d", fd, alnout);
    if  (!diment) {
        ex = -EX_USAGE;
        ifBLIN_QX0("diment==NULL");
        goto fout;
    }
    if  (!!(diment->flags & DIMENT_BYTES)) {
        alnout += 3;
    }
    if  (br.xdown < diment->margin.xdown) ex = -EDOM;
    if  (br.xup   > diment->margin.xup  ) ex = -EDOM;
    if  (br.ydown < diment->margin.ydown) ex = -EDOM;
    if  (br.yup   > diment->margin.yup  ) ex = -EDOM;
    br.xdown = MAX(br.xdown, diment->margin.xdown);
    br.xup   = MIN(br.xup  , diment->margin.xup  );
    br.ydown = MAX(br.ydown, diment->margin.ydown);
    br.yup   = MIN(br.yup  , diment->margin.yup  );
    if  (!!ex) {
        ifBLIN_QX3("out of margin");
        goto fout;
    }
    tmpfl = diment_doflag(diment, DIMENT_TKNENT, BLIN_CTL_FRST);
    if  (!px) isJW(-errno, "no mem");
    wsize = DIMENT_BITBIGCEIL(diment->unit, alnout);
    if  (diment->unit > wsize) isJX(-ENOTRECOVERABLE, "WRONG MATH");
    bzero(buf, 256);
    if  (wsize == diment->alnkoef) {
        //goto out; XXXX
    }
    if  (wsize % 8 == 0) {
        wsize /= 8;  /* wsize become bytes */
        for (int32_t y = br.ydown; y <= br.yup; y++) {
            for (int32_t x = br.xdown; x <= br.xup; x++) {
                ex = diment_un_tk(diment, x, y, px);
                if  (!!ex) isJX(ex, "diment_un_tk");
                psize = write(fd, px, DIMENT_b2B(diment->unit));
                if  (psize != DIMENT_b2B(diment->unit)) isJW(-EIO, "write");
                printed += psize;
                psize = write(fd, DIMENT_ZEROS256, wsize - DIMENT_b2B(diment->unit));
                if  (psize != wsize - DIMENT_b2B(diment->unit)) isJW(-EIO, "write");
                printed += psize;
            }
        }
    } else {
        if  (wsize >= 256 * 8) isJX(-ENOEXEC, "UNAVAILABLE");
        for (int32_t y = br.ydown; y <= br.yup; y++) {
            for (int32_t x = br.xdown; x <= br.xup; x++) {
                ex = diment_un_tk(diment, x, y, px);
                if  (!!ex) isJX(ex, "diment_un_tk");
                
                for (int i = 0; i < diment->unit; i++) {
                    buf[(bufsz + i) >> 3] |= ((px[i >> 3] >> (i & 7)) & 1) << ((bufsz + i) & 7);
                }
                /*
                word.i <<= bufi & 7;
                buff[(bufi >> 3)] |= word.c[0];
                for (int i = 1; i < 8; i++) {
                    buff[(bufi >> 3) + i] = word.c[i];
                }
                */
                bufsz += wsize;
                
                if  (bufsz + wsize > 256 * 8) {
                    psize = write(fd, buf, bufsz / 8);
                    if  (psize != bufsz / 8) isJW(-EIO, "write");
                    printed += psize;
                    buf[0] = buf[bufsz / 8];
                    bufsz &= 7;
                    bzero(&buf[1], 255);
                }
            }
            psize = write(fd, buf, DIMENT_b2B(bufsz));
            if  (psize != DIMENT_b2B(bufsz)) isJW(-EIO, "write");
            printed += psize;
            bufsz = 0;
            bzero(buf, 256);
        }
    }
out:
    diment_doflag(diment, tmpfl, BLIN_CTL_FEQU);
fout:
    free(px);
    ifBLIN_QX3("- %d %d", ex, printed);
    if  (!ex) {
        return(ex);
    } else {
        return(printed);
    }
#   undef isJX
#   undef isJW
#   undef blin_internal_flags
}

